/* TCTL v1.0
 * Tape ConTroL
 * This program is donated to the Public Domain
 *
 * If you have any suggestions or have found any bugs please write to:
 *  e-mail isj@aask.dk   or   fidonet 2:238/64.17
 * If you change this program and distribute it, please change the line
 * above to reflect YOUR address. I dont want to correct your bugs.
 */

/* Known bugs/limitations/missing features:
 * There is no 'format' or 'quick-erase' command
 * 'seek' and 'tell' is also missing
 * Variable-length blocks are not supported
 * I have only been able to test it with a SONY SDT-7000 DAT drive
 */

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <winioctl.h>
#include <stdio.h>
#include <stdlib.h>

int verbose=0;
static void Verbose(const char *fmt, ...) {
        if(verbose) {
                va_list ap;
                va_start(ap,fmt);
                vprintf(fmt,ap);
                va_end(ap);
        }
}

static void barf(int retcode, const char *fmt, ...) {
        //print error message and exit
        va_list ap;
        va_start(ap,fmt);
        fprintf(stderr,"tctl:");
        vfprintf(stderr,fmt,ap);
        va_end(ap);
        exit(retcode);
}

static const char *errorName(DWORD dw) {
        //Map a win32 error code to something understadable
        static struct {
                DWORD dw; const char *e;
        } errors[] = {
                {ERROR_BEGINNING_OF_MEDIA,      "Beginning of media"},
                {ERROR_BUS_RESET,               "Bus reset"},
                {ERROR_END_OF_MEDIA,            "End of media"},
                {ERROR_FILEMARK_DETECTED,       "Filemark detected"},
                {ERROR_SETMARK_DETECTED,        "Setmark detected"},
                {ERROR_NO_DATA_DETECTED,        "No data"},
                {ERROR_PARTITION_FAILURE,       "Partition failure/tape could not be partitioned"},
                {ERROR_INVALID_BLOCK_LENGTH,    "Invalid block length"},
                {ERROR_DEVICE_NOT_PARTITIONED,  "Device not partitioned"},
                {ERROR_MEDIA_CHANGED,           "Media may have changed"},
                {ERROR_NO_MEDIA_IN_DRIVE,       "No media in drive"},
                {ERROR_NOT_SUPPORTED,           "Not supported"},
                {ERROR_UNABLE_TO_LOCK_MEDIA,    "Could not lock media"},
                {ERROR_UNABLE_TO_UNLOAD_MEDIA,  "Could not unlock media"},
                {ERROR_WRITE_PROTECT,           "Media is write protected"},
                {0,0}
        };
        for(int i=0; errors[i].e; i++) {
                if(dw==errors[i].dw)
                        return errors[i].e;
        }
        static char errbuf[256];
        sprintf(errbuf,"error code=%ld",dw);
        return errbuf;
}

        
int got_tgdp=0; //Has we already retrieved drive parameters?
TAPE_GET_DRIVE_PARAMETERS tgdp;
static void GetDriveParameters(HANDLE hTape) {
        if(got_tgdp) return;
        DWORD dwSize=sizeof(tgdp);
        DWORD dw = GetTapeParameters(hTape, GET_TAPE_DRIVE_INFORMATION, &dwSize, &tgdp);
        if(dw!=NO_ERROR)
                barf(2,"Could not retrieve drive parameters, %s",errorName(dw));
        got_tgdp=1;
}


static void tape_setcompress(HANDLE hTape, int onoff) {
        Verbose("setcompress %s\n",onoff?"on":"off");
        
        GetDriveParameters(hTape);
        if((tgdp.FeaturesLow&TAPE_DRIVE_COMPRESSION)==0 && onoff)
                barf(1,"Cannot enable compression, tape does not support compression");
        TAPE_SET_DRIVE_PARAMETERS tsdp;
        tsdp.ECC = tgdp.ECC;
        tsdp.Compression = (BOOLEAN)onoff;
        tsdp.DataPadding = tgdp.DataPadding;
        tsdp.ReportSetmarks = tgdp.ReportSetmarks;
        tsdp.EOTWarningZoneSize = tgdp.EOTWarningZoneSize;

        DWORD dw = SetTapeParameters(hTape, SET_TAPE_DRIVE_INFORMATION, &tsdp);
        if(dw!=NO_ERROR)
                barf(2,"Could not change compression, %s",errorName(dw));
}


static void tape_eof(HANDLE hTape, int count) {
        Verbose("weof %d",count);
        GetDriveParameters(hTape);
        DWORD tapemarkType;
        if(tgdp.FeaturesHigh&TAPE_DRIVE_WRITE_FILEMARKS) {
                tapemarkType = TAPE_FILEMARKS;
                Verbose(" (normal filemark)\n");
        } else if(tgdp.FeaturesHigh&TAPE_DRIVE_WRITE_LONG_FMKS) {
                tapemarkType = TAPE_LONG_FILEMARKS;
                Verbose(" (long filemark)\n");
        } else if(tgdp.FeaturesHigh&TAPE_DRIVE_WRITE_SHORT_FMKS) {
                tapemarkType = TAPE_SHORT_FILEMARKS;
                Verbose(" (short filemark)\n");
        } else
                barf(1,"Drive does not support filemarks");
        DWORD dw=WriteTapemark(hTape,tapemarkType,count,FALSE);
        if(dw!=NO_ERROR)
                barf(2,"Could not write mark, %s",errorName(dw));
}

static void tape_fsf(HANDLE hTape, int count) {
        Verbose("fsf %d\n",count);
        GetDriveParameters(hTape);
        DWORD dw=SetTapePosition(hTape, TAPE_SPACE_FILEMARKS, 0, count, 0, FALSE);
        if(dw!=NO_ERROR)
                barf(2,"Could not forward-scan filemark, %s",errorName(dw));
}

static void tape_bsf(HANDLE hTape, int count) {
        Verbose("bsf %d\n",count);
        GetDriveParameters(hTape);
        DWORD dw=SetTapePosition(hTape, TAPE_SPACE_FILEMARKS, 0, -count, 0xffffffff, FALSE);
        if(dw!=NO_ERROR)
                barf(2,"Could not backward-scan filemark, %s",errorName(dw));
}

static void tape_fsr(HANDLE hTape, int count) {
        Verbose("fsr %d\n",count);
        GetDriveParameters(hTape);
        DWORD dw=SetTapePosition(hTape, TAPE_SPACE_RELATIVE_BLOCKS, 0, count, 0, FALSE);
        if(dw!=NO_ERROR)
                barf(2,"Could not forward-scan blocks, %s",errorName(dw));
}

static void tape_bsr(HANDLE hTape, int count) {
        Verbose("bsr %d\n",count);
        GetDriveParameters(hTape);
        DWORD dw=SetTapePosition(hTape, TAPE_SPACE_RELATIVE_BLOCKS, 0, -count, 0xffffffff, FALSE);
        if(dw!=NO_ERROR)
                barf(2,"Could not backward-scan blocks, %s",errorName(dw));
}


static void tape_rewind(HANDLE hTape, int) {
        Verbose("rewind\n");
        DWORD dw=SetTapePosition(hTape, TAPE_REWIND, 0,0,0, FALSE);
        if(dw!=NO_ERROR)
                barf(2,"Could not rewind, %s",errorName(dw));
}

static void tape_offline(HANDLE hTape, int) {
        Verbose("offline\n");
        //the SDK doc says that PrepareTape(...TAPE_UNLOAD) rewinds the tape
        //We don't want that to happen, since that what the 'rewoffl' command is fore
        //We have to use IOCTL and hope that doesn't rewind the tape :-(
        BOOL b;
        DWORD notUsed;
        b = DeviceIoControl(hTape, IOCTL_DISK_EJECT_MEDIA, NULL,0, NULL,0, &notUsed, NULL);
        if(!b)
                barf(2,"Could not take tape offline, %s",errorName(GetLastError()));
}

static void tape_online(HANDLE hTape, int) {
        Verbose("online");
        //the SDK doc says that PrepareTape(...TAPE_LOAD) rewinds the tape
        //We don't want that to happen, since that what the 'onlrew' command is fore
        //We have to use IOCTL :-(
        BOOL b;
        DWORD notUsed;
        b = DeviceIoControl(hTape, IOCTL_DISK_LOAD_MEDIA, NULL,0, NULL,0, &notUsed, NULL);
        if(!b)
                barf(2,"Could not put tape online, %s",errorName(GetLastError()));
}

static void tape_rewoffl(HANDLE hTape, int) {
        Verbose("rewoffl\n");
        DWORD dw=PrepareTape(hTape, TAPE_UNLOAD, FALSE);
        if(dw!=NO_ERROR)
                barf(1,"Could not rewoffl, %s",errorName(dw));
}

static void tape_onlrew(HANDLE hTape, int) {
        Verbose("onlrew\n");
        DWORD dw=PrepareTape(hTape, TAPE_LOAD, FALSE);
        if(dw!=NO_ERROR)
                barf(2,"Could not onlrew, %s",errorName(dw));
}

static void tape_reserve(HANDLE hTape, int) {
        Verbose("reserve\n");
        DWORD dw=PrepareTape(hTape, TAPE_LOCK, FALSE);
        if(dw!=NO_ERROR)
                barf(2,"Could not lock/reserve, %s",errorName(dw));
}

static void tape_release(HANDLE hTape, int) {
        Verbose("release\n");
        DWORD dw=PrepareTape(hTape, TAPE_UNLOCK, FALSE);
        if(dw!=NO_ERROR)
                barf(2,"Could not release/unlock, %s",errorName(dw));
}

static void tape_erase(HANDLE hTape, int) {
        Verbose("erase\n");
        //Since the erase commands may or may not erase from the current
        //position we have to rewind the tape first.
        
        //first, rewind the tape
        DWORD dw=SetTapePosition(hTape, TAPE_REWIND, 0,0,0, FALSE);
        if(dw!=NO_ERROR)
                barf(2,"Could not rewind, %s",errorName(dw));
        //then erase from current position (=start) to the end
        dw = EraseTape(hTape,TAPE_ERASE_SHORT, FALSE);
        if(dw!=NO_ERROR) //error? if so try the long erase
                dw = EraseTape(hTape,TAPE_ERASE_LONG, FALSE);
        if(dw!=NO_ERROR)
                barf(2,"Could not erase tape, %s",errorName(dw));
}  

static void tape_retension(HANDLE hTape, int) {
        Verbose("retension\n");
        DWORD dw = PrepareTape(hTape, TAPE_TENSION, FALSE);
        if(dw!=NO_ERROR)
                barf(2,"Could not retension, %s",errorName(dw));
}

char deviceName[256];

static void tape_status(HANDLE hTape, int) {
        GetDriveParameters(hTape);
        printf("Status for drive %s:\n",deviceName);
        printf("  ECC                : %s\n", tgdp.ECC?"Supported":"Not supported");
        printf("  Compression        : %s\n", tgdp.Compression?"Enabled":"Disabled");
        printf("  Default block size : %ld\n", tgdp.DefaultBlockSize);
        printf("  Minimum block size : %ld\n", tgdp.MinimumBlockSize);
        printf("  Maximum block size : %ld\n", tgdp.MaximumBlockSize);
        printf("  Features: \n");
        if(tgdp.FeaturesLow&TAPE_DRIVE_COMPRESSION) printf("    Compression supported\n");
        if(tgdp.FeaturesLow&TAPE_DRIVE_GET_ABSOLUTE_BLK) printf("    Absolute block addressing\n");
        if(tgdp.FeaturesLow&TAPE_DRIVE_GET_LOGICAL_BLK) printf("    Logical block addressing\n");
        if(tgdp.FeaturesLow&TAPE_DRIVE_TAPE_CAPACITY) printf("    Capacity reporting\n");
        if(tgdp.FeaturesLow&TAPE_DRIVE_TAPE_REMAINING) printf("    Remaining capacity reporting\n");
        if(tgdp.FeaturesLow&TAPE_DRIVE_VARIABLE_BLOCK) printf("    Variable block sizes\n");
        if(tgdp.FeaturesHigh&TAPE_DRIVE_LOCK_UNLOCK) printf("    lock/unlock supported\n");

        printf("Media status:\n");
        TAPE_GET_MEDIA_PARAMETERS tgmp;
        DWORD dwSize=sizeof(tgmp);
        DWORD dw = GetTapeParameters(hTape, GET_TAPE_MEDIA_INFORMATION, &dwSize, &tgmp);
        if(dw==ERROR_NO_MEDIA_IN_DRIVE) {
                printf("  No media in drive\n");
                return;
        } else if(dw!=NO_ERROR)
                barf(1,"Could not retrieve tape information, %s",errorName(dw));

        if(tgdp.FeaturesLow&TAPE_DRIVE_TAPE_CAPACITY) {
                double cap = tgmp.Capacity.HighPart*4294967296.0+tgmp.Capacity.LowPart;
                printf("  Capacity        : %12.0f bytes (~%ldMB) (0x%08x%08x)\n",
                       cap,
                       (long)(cap/1024/1024),
                       tgmp.Capacity.HighPart,tgmp.Capacity.LowPart);
        }
        if(tgdp.FeaturesLow&TAPE_DRIVE_TAPE_REMAINING) {
                double rem = tgmp.Remaining.HighPart*4294967296.0+tgmp.Remaining.LowPart;
                printf("  Remaning        : %12.0f bytes (~%ldMB) (0x%08x%08x)\n",
                       rem,
                       (long)(rem/1024/1024),
                       tgmp.Remaining.HighPart,tgmp.Remaining.LowPart);
        }
        if(tgdp.FeaturesLow&(TAPE_DRIVE_FIXED|TAPE_DRIVE_INITIATOR))
                printf("  Partitions      : %ld\n",tgmp.PartitionCount);
        printf("  Block size      : %ld bytes\n",tgmp.BlockSize);
        printf("  Write protected : %s\n", tgmp.WriteProtected?"Yes":"No");
}



static void tape_read(HANDLE hTape, HANDLE hf, int BlockSize) {
        Verbose("read\n");
        if(BlockSize==-1) {
                TAPE_GET_MEDIA_PARAMETERS tgmp;
                DWORD dwSize=sizeof(tgmp);
                DWORD dw = GetTapeParameters(hTape, GET_TAPE_MEDIA_INFORMATION, &dwSize, &tgmp);
                if(dw!=NO_ERROR)
                        barf(2,"Could not retrieve media parameters, %s",errorName(dw));
                BlockSize = tgmp.BlockSize;
        }
        //don't use a silly small buffer
        DWORD bufferSize = BlockSize;
        while(bufferSize<32768) bufferSize<<=1;
        
        void *buf=malloc(bufferSize);
        if(!buf)
                barf(3,"Could not allocate memory for buffer");
        
        for(;;) {
                DWORD bytesRead=0;
                BOOL read_b = ReadFile(hTape, buf, bufferSize, &bytesRead, NULL);
                if((!read_b) && bytesRead==0) {
                        DWORD dw=GetLastError();
                        if(dw==ERROR_FILEMARK_DETECTED ||
                           dw==ERROR_NO_DATA_DETECTED)
                                break; //These two errors are OK
                        barf(2,"Could not read from tape, %s",errorName(dw));
                }
                DWORD bytesWritten;
                BOOL write_b = WriteFile(hf, buf, bytesRead, &bytesWritten, NULL);
                if((!write_b) || bytesWritten!=bytesRead)
                        barf(3,"Could not write to file, %s",errorName(GetLastError()));
                if(!read_b) break;
        }
        free(buf);
}

static void tape_write(HANDLE hTape, HANDLE hf, int BlockSize) {
        Verbose("write\n");
        if(BlockSize==-1) {
                //Ask the tape what size the blocks are
                TAPE_GET_MEDIA_PARAMETERS tgmp;
                DWORD dwSize=sizeof(tgmp);
                DWORD dw = GetTapeParameters(hTape, GET_TAPE_MEDIA_INFORMATION, &dwSize, &tgmp);
                if(dw!=NO_ERROR)
                        barf(2,"Could not retrieve media parameters, %s",errorName(dw));
                BlockSize = tgmp.BlockSize;
        }
        
        //don't use a silly small buffer
        DWORD bufferSize = BlockSize;
        while(bufferSize<32768) bufferSize<<=1;

        void *buf=malloc(bufferSize);
        if(!buf)
                barf(3,"Could not allocate memory for buffer");
        
        for(;;) {
                DWORD bytesRead;
                BOOL b = ReadFile(hf, buf, bufferSize, &bytesRead, NULL);
                if(!b) {
                        barf(3,"Could not read from file, %s",errorName(GetLastError()));
                }
                DWORD bytesToWrite;
                if(bytesRead<bufferSize) {
                        //round up - zerofilling
                        memset(((char*)buf)+bytesRead, 0, bufferSize-bytesRead);
                        bytesToWrite = bytesRead + (BlockSize - bytesRead % BlockSize);
                } else
                        bytesToWrite = bytesRead;
                DWORD bytesWritten;
                b = WriteFile(hTape, buf, bytesToWrite, &bytesWritten, NULL);
                if((!b) || bytesWritten!=bytesToWrite)
                        barf(3,"Could not write to tape, %s",errorName(GetLastError()));
                        if(bytesRead<bufferSize) break;
        }
        free(buf);
}



static void findDeviceName() {
        //Figure out what the name of the tape device is
        
        //TAPE environment variable set?
        char *e=getenv("TAPE");
        if(e) {
                strcpy(deviceName,e);
                return;
        }
        //hmmm... the default device is usually \\.\tape0
        //But to be sure we really ought to traverse the registry under
        //\\hkey_local_machine\hardware\devicemap\[usually:scsi]...
        //and find a subkey with a value named 'type' with the value
        //'TapePeripheral'
        
        //But this is much easier:
        strcpy(deviceName,"\\\\.\\TAPE0");
}

int main(int argc, char **argv) {
        if(argc<2) {
                printf("tctl v1.0 by ISJ (compiled:"__DATE__" "__TIME__")\n"
                       "usage:\n"
                       "tctl [-f Device] [-v] [-b BlockSize] [eof|weof|fsf|bsf|fsr|bsr|rewind|offline|rewoffl|erase|retension|reset|status|read File|write File] Count\n"
                       "Use tctl -usage for explanation\n"
                      );
                return 1;
        }
        if(stricmp(argv[1],"-usage")==0 ||
           stricmp(argv[1],"-?")==0)
        {
                printf("tctl v1.0 by ISJ (compiled:"__DATE__" "__TIME__")\n"
                       "Issue commands to tape device\n"
                       "usage:\n"
                       "tctl [-f Device] [-v] [-b BlockSize] command [command...]\n"
                       "  -f Device     Specify device name. Default is \\\\.\\TAPE0\n"
                       "  -v            Verbose\n"
                       "  -b BlockSize  Specify block size. The default is the blocksize of the tape\n"
                       "Tape & misc commands:\n"
                       "  offline, unload\n"
                       "                Take tape offline (=unload).\n"
                       "  online, load  Put the tape online (load). Can cause retension.\n"
                       "  retension     Retension tape. Usually causes the tape to move to the end then\n"
                       "                to the start.\n"
                       "  rewoffl       Rewind tape and take if offline\n"
                       "  onlrew        Put tape online and rewind it.\n"
                       "  reset         Ignored.\n"
                       "  status        Reports status of tape drive and tape (if any).\n"
                       //reserve
                       //release
                       "Head movement commands:\n"
                       "  fsf [Count]   Scans Count filemarks forward\n"
                       "  bsf [Count]   Scans Count filemarks backward\n"
                       "  fsr [Count]   Scans Count records forward\n"
                       "  bsr [Count]   Scans Count records backward\n"
                       "  rewind        Rewind tape\n"
                       "Read/Write commands:\n"
                       "  eof, weof [Count]\n"
                       "                Write eof mark (filemark)\n"
                       "  erase         Erase entire tape\n"
                       "  read [File]   Reads from tape until filemark/eof/setmark/EOT is found and\n"
                       "                writes to File\n"
                       "  write [File]  Reads File and writes it to tape (no filemark is written)\n"
                       "Note:\n"
                       "  The first command on a newly inserted tape will usually fail on SCSI devices\n"
                       "Examples:\n"
                       "  Rewind tape:\n"
                       "    tctl rewind\n"
                       "  Rewind the tape, write the file 'myfile.ext' to the tape, write a filemark,\n"
                       "  rewind the tape and eject it:\n"
                       "    tctl rewind write myfile.ext weof rewoffl\n"
                       "  Read second file from tape:\n"
                       "    tctl rewind fsf read somefile.ext\n"
                       "Exit codes:\n"
                       "  0  Success\n"
                       "  1  Command line error\n"
                       "  2  Tape error\n"
                       "  3  non-tape error\n"
                      );
        }
        
        findDeviceName();

        HANDLE hTape;
        hTape  = CreateFile(deviceName,
                            GENERIC_READ|GENERIC_WRITE,
                            0,
                            0,
                            OPEN_EXISTING,
                            0,
                            NULL
                           );
        if(hTape==INVALID_HANDLE_VALUE)
                barf(1,"Could not open tape device '%s'\n");

        int BlockSize=-1;
        for(int a=1; a<argc; a++) {
                if(argv[a][0]=='-') {
                        //parse option
                        if(strcmp(argv[a],"-f")==0) {
                                if(a<argc-1) {
                                        strcpy(deviceName,argv[a+1]);
                                        a++;
                                } else {
                                        barf(1,"-f parameter requires a device name");
                                }
                        } else if(strcmp(argv[a],"-v")==0) {
                                verbose=1;
                        } else if(strcmp(argv[a],"-b")==0) {
                                if(a<argc-1) {
                                        BlockSize=atoi(argv[a+1]);
                                        if(BlockSize<=0)
                                                barf(1,"-b parameter is invalid");
                                } else
                                        barf(1,"-b option needs a parameter.");
                                a++;
                        } else {
                                barf(1,"Unknown option: %s",argv[a]);
                        }
                } else {
                        //parse command
                        const char *cmd=argv[a];
                        int count=1;
                        if(a<argc-1 &&
                           argv[a+1][0]>='0' && argv[a+1][0]<='9' &&
                           strcmp(cmd,"read")!=0 && strcmp(cmd,"write")!=0)
                        {
                                //commands except read and write may take an option count parameter
                                char *endptr;
                                count=strtol(argv[a+1],&endptr,0);
                                if(count<0 || (endptr && *endptr))
                                        barf(1,"error in count argument '%s' to command '%s'",argv[a+1],cmd);
                                a++;
                        } else
                                count=1;
                        if(strcmp(cmd,"compress")==0) {
                                if(a>=argc-1 || (strcmp(argv[a+1],"on")==0 && strcmp(argv[a+1],"off")==0))
                                        barf(1,"compress commands take one argument: 'on' or 'off'");
                                tape_setcompress(hTape,strcmp(argv[a+1],"on")==0);
                                a++;
                        } else if(strcmp(cmd,"eof")==0 || strcmp(cmd,"weof")==0) {
                                tape_eof(hTape,count);
                        } else if(strcmp(cmd,"fsf")==0) {
                                tape_fsf(hTape,count);
                        } else if(strcmp(cmd,"bsf")==0) {
                                tape_bsf(hTape,count);
                        } else if(strcmp(cmd,"fsr")==0) {
                                tape_fsr(hTape,count);
                        } else if(strcmp(cmd,"bsr")==0) {
                                tape_bsr(hTape,count);
                        } else if(strcmp(cmd,"rewind")==0) {
                                tape_rewind(hTape,count);
                        } else if(strcmp(cmd,"offline")==0 || strcmp(cmd,"unload")==0) {
                                tape_offline(hTape,count);
                        } else if(strcmp(cmd,"online")==0 || strcmp(cmd,"load")==0) {
                                tape_online(hTape,count);
                        } else if(strcmp(cmd,"rewoffl")==0) {
                                tape_rewoffl(hTape,count);
                        } else if(strcmp(cmd,"onlrew")==0) {
                                tape_onlrew(hTape,count);
                        } else if(strcmp(cmd,"reserve")==0 || strcmp(cmd,"lock")==0) {
                                tape_reserve(hTape,count);
                        } else if(strcmp(cmd,"release")==0 || strcmp(cmd,"unlock")==0) {
                                tape_release(hTape,count);
                        } else if(strcmp(cmd,"erase")==0) {
                                tape_erase(hTape,count);
                        } else if(strcmp(cmd,"retension")==0) {
                                tape_retension(hTape,count);
                        } else if(strcmp(cmd,"reset")==0) {
                                //ignored
                        } else if(strcmp(cmd,"status")==0) {
                                tape_status(hTape,count);
                        } else if(strcmp(cmd,"read")==0) {
                                HANDLE hf;
                                if(a<argc-1) {
                                        //write to file
                                        hf = CreateFile(argv[a+1], GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL);
                                        if(hf==INVALID_HANDLE_VALUE)
                                                barf(3,"Could not create '%s', ",argv[a+1],errorName(GetLastError()));
                                        tape_read(hTape,hf,BlockSize);
                                        a++;
                                        CloseHandle(hf);
                                } else {
                                        //write to stdout
                                        hf = GetStdHandle(STD_OUTPUT_HANDLE);
                                        tape_read(hTape,hf,BlockSize);
                                }
                        } else if(strcmp(cmd,"write")==0) {
                                HANDLE hf;
                                if(a<argc-1) {
                                        //read from file
                                        hf = CreateFile(argv[a+1], GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
                                        if(hf==INVALID_HANDLE_VALUE)
                                                barf(3,"Could not open '%s', ",argv[a+1],errorName(GetLastError()));
                                        tape_write(hTape,hf,BlockSize);
                                        a++;
                                        CloseHandle(hf);
                                } else {
                                        //read from stdin
                                        hf = GetStdHandle(STD_INPUT_HANDLE);
                                        tape_write(hTape,hf,BlockSize);
                                }
                        } else {
                                barf(1,"Unknown command: '%s'",cmd);
                        }
                }
        }

        CloseHandle(hTape);
        
        return 0;        
}

